from ..error.wsgi import WsgiResBodyError
from ..error.http import code_to_str
from ..iface import ISendHandler


class SendHandler(ISendHandler):
    def handle(self, req, res, wsgi_send):
        header = [(k, v) for k, v in res.header.items()]
        if res.content_type is not None:
            header.append(('Content-Type', res.content_type))
        if res.body is None:
            return []
        elif not isinstance(res.body, bytes):
            raise WsgiResBodyError(res.body)
        else:
            wsgi_send(code_to_str(res.status), header)
            return [res.body]



'''
SYNOPSIS

    interface IRecvHandler

DESCRIPTION

    Interface to initiallize request and response in WSGI application.
'''

from abc import ABC, abstractmethod


class IRecvHandler(ABC):
    @abstractmethod
    def handle(self, req, res, wsgi_env):
        pass



'''
SYNOPSIS

    interface ISendHandler

DESCRIPTION

    Interface to send response message to client in WSGI application.
'''

from abc import ABC, abstractmethod


class ISendHandler(ABC):
    @abstractmethod
    def handle(self, req, res, wsgi_send):
        pass



from ..iface import IPipeHandler


class ResCorsHandler(IPipeHandler):
    def handle(self, req, res):
        if req.method.lower() != 'option':
            return
        res.header['Access-Control-Allow-Origin'] = '*'
        res.header['Access-Control-Allow-Methods'] = \
            'GET,POST,PUT,DELETE,OPTIONS'
        res.header['Access-Control-Allow-Headers'] = \
            'Authorization,Content-Type'



'''
SYNOPSIS

    class Route

DESCRIPTION

    Decorate function to add it's function to routing.
'''

import os

from ..mime.type import MIME_JSON
from ..etc import URL_SLASH
from .error import RoutePathError
from .type import RouteSpec


class Route():
    def __init__(self, base_path):
        if not self.verify_path(base_path):
            raise RoutePathError(base_path)
        self.base_path = base_path
        self.specs = []

    def get(self, path):
        return self.map('get', path, None)

    def post(self, path, req_type=MIME_JSON):
        return self.map('post', path, req_type)

    def put(self, path, req_type=MIME_JSON):
        return self.map('put', path, req_type)

    def patch(self, path, req_type=MIME_JSON):
        return self.map('patch', path, req_type)

    def delete(self, path):
        return self.map('delete', path, None)

    def option(self, path):
        return self.map('option', path, None)

    def map(self, method, path, req_type):
        if not self.verify_path(path):
            raise RoutePathError(path)

        def decorator_fn(target_fn):
            abs_path = os.path.join(self.base_path, path)
            spec = RouteSpec(method, abs_path, target_fn, req_type)
            self.specs.append(spec)
        return decorator_fn

    def verify_path(self, path):
        if len(path) > 0:
            if path[0] == URL_SLASH:
                return False
        return True



from .recv import *
from .send import *

from .req_json import *
from .req_log import *
from .req_url_encode import *

from .res_cors import *
from .res_json import *

from .err_log import *
from .err_http import *



'''
SYNOPSIS

    interface IRouter

DESCRIPTION

    Interface to find handler from request.

    Detect request handle and url parameters. It's not support parameters in
    url for performace, let use arugments.
'''

from abc import ABC, abstractmethod


class IRouter(ABC):
    @abstractmethod
    def find_handle(self, req):
        pass



from urllib.parse import parse_qsl

from ..error.http import Http400Error
from ..iface import IPipeHandler
from ..mime.type import MIME_URLENCODE


class ReqUrlEncodeHandler(IPipeHandler):
    def handle(self, req, res):
        if req.content_type != MIME_URLENCODE:
            return
        if req.body is None:
            return
        try:
            values = dict(parse_qsl(req.body.decode('utf-8')))
            if req.body is not None and len(values) == 0:
                raise Http400Error(req, 'body is invaild url encode')
            req.body = values
        except ValueError:
            raise Http400Error(req)



import json

from ..error.http import Http400Error
from ..iface import IPipeHandler
from ..etc import UTF_8
from ..mime.type import MIME_JSON


class ReqJsonHandler(IPipeHandler):
    def handle(self, req, res):
        if req.content_type != MIME_JSON:
            return
        if req.body is None:
            return
        try:
            req.body = json.loads(req.body.decode(UTF_8))
        except ValueError:
            raise Http400Error(req, 'body is invalid json format')



class Request():
    method = None
    path = None
    query_str = None
    content_type = None
    content_length = None
    server_name = None
    server_port = None
    server_protocol = None
    remote_addr = None
    remote_port = None

    header = {}
    body = None

    args = {}

    def __init__(self):
        pass



'''
DESCRIPTION

    Miscellaneous things is too small to classify.
'''

from stat import S_IRUSR, S_IWUSR, S_IRGRP, S_IWGRP, S_IROTH

URL_SLASH = '/'
UTF_8 = 'utf-8'
LOGFILE_MODE = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH



'''
SYNOPSIS

    interface IErrorHandler

DESCRIPTION

    Interface to handle error in pipe line processing.
'''

from abc import ABC, abstractmethod


class IErrorHandler(ABC):
    @abstractmethod
    def handle(self, req, res, e):
        pass



'''
SYNOPSIS

    class RouteExistError
    class RoutePathError

DESCRIPTION

    Error occurs during routing.
'''


class RouteExistError(Exception):
    def __init__(self, spec):
        self.spec = spec
        self._msg = '%s %s; %s' % (
            spec.method, spec.path, spec.req_type
        )

    def __str___(self):
        return self._msg


class RoutePathError(Exception):
    def __init__(self, path):
        self.path = path

    def __str___(self):
        return self.path



import json

from ..iface import IPipeHandler
from ..mime.type import MIME_JSON
from ..etc import UTF_8


class ResJsonHandler(IPipeHandler):
    def handle(self, req, res):
        if res.content_type != MIME_JSON:
            return
        if res.body is None:
            return
        res.body = json.dumps(res.body).encode(UTF_8)



import logging
from os.path import isfile

from ..etc import LOGFILE_MODE
from ..iface import IPipeHandler
from ..shell import touch


class ReqLogHandler(IPipeHandler):
    def __init__(self, file):
        self._file = file
        self.logger = logging.getLogger(file)

        if not isfile(file):
            touch(file, LOGFILE_MODE)

        fhandler = logging.FileHandler(file)
        formatter = logging.Formatter('%(asctime)s$ %(message)s')
        fhandler.setFormatter(formatter)
        self.logger.addHandler(fhandler)
        self.logger.setLevel(logging.INFO)

    @property
    def file(self):
        return self._file

    def handle(self, req, res):
        msg = '%s %s %s' % (req.remote_addr, req.method, req.path)
        self.logger.info(msg)



"""
SYNOPSIS

    class Router

DESCRIPTION

    Implement IRouter.
"""

from ..error.http import Http404Error, Http405Error, Http406Error
from ..iface import IRouter
from ..etc import URL_SLASH
from .error import RouteExistError
from .type import RouteNode


class Router(IRouter):
    def __init__(self, routes):
        self._root_route = RouteNode('/')
        for route in routes:
            for spec in route.specs:
                self.add_route(spec)

    def find_handle(self, req):
        node = self.find_node(req.path)
        if node is None or len(node.specs) == 0:
            raise Http404Error(req)
        m_ok_specs = [s for s in node.specs if s.method == req.method.lower()]
        if len(m_ok_specs) == 0:
            raise Http405Error(req)
        for spec in m_ok_specs:
            if spec.req_type == req.content_type:
                return spec.handler
        raise Http406Error(req)

    def add_route(self, spec):
        # search and add nodes
        node = self._root_route
        if spec.path != URL_SLASH:
            node_names = spec.path.split(URL_SLASH)
            node_names = list(filter(''.__ne__, node_names))
            for node_name in node_names:
                if node_name not in node.child:
                    node.child[node_name] = RouteNode(node_name)
                node = node.child[node_name]

        # add handle
        if self._spec_is_exist(node, spec):
            raise RouteExistError(spec)
        node.specs.append(spec)

    def find_node(self, path):
        node = self._root_route
        if path != URL_SLASH:
            node_names = path.split(URL_SLASH)
            node_names = list(filter(''.__ne__, node_names))
            for node_name in node_names:
                if node_name not in node.child:
                    return None
                node = node.child[node_name]
        return node

    def _spec_is_match(self, spec_1, spec_2):
        if spec_1.method != spec_2.method:
            return False
        if spec_1.req_type != spec_2.req_type:
            return False
        return True

    def _spec_is_exist(self, node, spec):
        for node_spec in node.specs:
            if self._spec_is_match(node_spec, spec):
                return True
        return False



'''
SYNOPSIS

    class Application

DESCRIPTION

    Implement application with WSGI interface.
'''

from os import path

from .iface import IWsgi
from .type import Request, Response
from .handler import RecvHandler, SendHandler
from .handler import ReqJsonHandler, ReqUrlEncodeHandler, ReqLogHandler
from .handler import ResJsonHandler, ResCorsHandler
from .handler import ErrorHttpHandler, ErrorLogHandler


class Application(IWsgi):
    req_log_handler = None
    err_log_handler = None

    err_handlers = None
    req_handlers = None
    res_handlers = None

    def __init__(self, name, router):
        self._name = name
        self._router = router

        self.recv_handler = RecvHandler()
        self.send_handler = SendHandler()

        req_logfile = path.join('/var/tmp', name, 'request.log')
        self.req_log_handler = ReqLogHandler(req_logfile)

        err_logfile = path.join('/var/tmp', name, 'error.log')
        self.err_log_handler = ErrorLogHandler(err_logfile)

        self.err_handlers = [ErrorHttpHandler()]
        self.req_handlers = [ReqJsonHandler(), ReqUrlEncodeHandler()]
        self.res_handlers = [ResJsonHandler(), ResCorsHandler()]

    def __call__(self, wsgi_env, wsgi_send):
        req = Request()
        res = Response()

        try:
            # receive message, initial req and res
            self.recv_handler.handle(req, res, wsgi_env)

            # perform request logging handler
            self.req_log_handler.handle(req, res)

            # routing, find main handler
            m_req_handle = self._router.find_handle(req)

            # perform request handlers
            for handler in self.req_handlers:
                handler.handle(req, res)

            # perform main handler
            m_req_handle(req, res)

            # perform response handlers
            for handler in self.res_handlers:
                handler.handle(req, res)

            # send response
            return self.send_handler.handle(req, res, wsgi_send)
        except Exception as e:
            # perform error handlers
            handled = False
            for handler in self.err_handlers:
                if handler.handle(req, res, e):
                    handled = True
            if not handled:
                raise e

            # perform error logging handler
            self.err_log_handler.handle(req, res, e)

            # send error response
            return self.send_handler.handle(req, res, wsgi_send)



'''
SYNOPSIS

    interface IPipeHandler

DESCRIPTION

    Inteface to process pipe processing.
'''

from abc import ABC, abstractmethod


class IPipeHandler(ABC):
    @abstractmethod
    def handle(self, req, res):
        pass



'''
SYNOPSIS
    class RouteNode
    class RouteSpec

DESCRIPTION

    Data structures are uses during routing.
'''


class RouteNode():
    def __init__(self, name):
        self.name = name
        self.child = {}
        self.specs = []


class RouteSpec():
    def __init__(self, method, path, handler, req_type):
        self.method = method
        self.path = path
        self.handler = handler
        self.req_type = req_type



'''
SYNOPSIS

    interface IWsgi

DESCRIPTION

    HTTP Server with WSGI interface. Entry is method __call__().

REFERENCES

    WSGI - Wikipedia
        https://en.wikipedia.org/wiki/Web_Server_Gateway_Interface

    PEP333 - Python Web Server Gateway Interface
        https://www.python.org/dev/peps/pep-0333/
'''

from abc import ABC, abstractmethod


class IWsgi(ABC):
    @abstractmethod
    def __call__(self, wsgi_env, wsgi_send):
        pass
